/* 
   Generator of monotone sequences of priority queue operations.
   Written by Andrew V. Goldberg, NEC Research Inst.,  avg@research.nj.nec.com.
   For format details, see the 5th DIMACS Implementation Challenge info.

*/
 
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <values.h>

#include "random.h"
#include "fprintfll.h"

#define RANGE 100
#define DASH '-'
#define OUT -1

/* types */
/*
  use long long int time to prevent overflow
  if long long not supported, use double
*/
typedef long long int timeType;

typedef struct itemSt
{
  long        heap_pos;  /* position in the heap */
  timeType    key;
  long        id;
} itemT;

typedef struct heapSt
{
  long    size;
  itemT **item;
} heap;

#include "heap.h"

/* macros */
#define checkIfDone {if (N <= 0) exit(0);}

void error (int errorNo)

{
  switch ( errorNo ) {

  case 1: {
    fprintf ( stderr, "\nUsage: mongen C K E N [-sS]\n");
    fprintf ( stderr, "      C is an upper bound on event duration \n");
    fprintf ( stderr, "      K is an approximate queue size\n");
    fprintf ( stderr, "      E determines how clustered the values are\n");
    fprintf ( stderr, "      N is the total number of operations\n");
    fprintf ( stderr, "      S is a seed\n");
    break;
  }
  case 2: {
    fprintf ( stderr, "max duration C must be 2 or greater\n");
    break;
  }
  case 3: {
    fprintf ( stderr, "approx size K must be positive\n");
    break;
  }
  case 4: {
    fprintf ( stderr, "Cluster parameter E must satisfy 0 < E <= C/3\n");
    break;
  }
  case 5: {
    fprintf ( stderr, "number of operations N must be nonnegative\n");
    break;
  }
  }

  exit(errorNo);
}

void main (int argc, char* argv[])
{

  char args[30];
  long C, E, K, N, step;
  long maxIndex;
  long i, j;
  long inc;
  long seed=31415;
  long curIndex=0;
  timeType T, curT;
  itemT *anItem, *nextItem;
  heap h;

  if ((argc < 5) || (argc > 6)) error(1);

  /* first parameter -- C */
  strcpy (args, argv[1]);
  if (( C = atoi(argv[1])) < 2) error (2);
 
  /* second parameter -- K */
  strcpy (args, argv[2]);
  if (( K = atoi(argv[2])) < 1) error (3);

  /* third parameter -- E */
  strcpy (args, argv[3]);
  if ((( E = atoi(argv[3])) < 0) || (E > C/3)) error (4);
 
  /* fourth parameter -- N */
  strcpy (args, argv[4]);
  if (( N = atoi(argv[4])) < 0) error (5);

  if ( argc == 6 ) {
    strcpy ( args, argv[5]);
    if (( args[0] != DASH ) || ( args[1] != 's')) error (1);
    seed  =  atoi ( &args[2] );
  }

  /* initialization */
  step = C - E;
  SetRandom(seed);
  makeHeap(h, K+1);
  curT = 0;
  T = step;

  /* figure out the number of insert operations, upper bound */
  maxIndex = K + (N/2) + 1;
  if (maxIndex > N) maxIndex = N;

  printf("pqh %d -M%d -c%d -C%d\n", maxIndex, K+1, 0, C);
  printf("com evsimgen parameters C=%d K=%d E=%d N=%d\n",
	 C, K, E, N);

  checkIfDone;

  /* insert K elements */
  for (i = 0; i < K; i++) {
    
    j = Random(0,E);
    curIndex++;
    anItem = (itemT *) calloc(1, sizeof(itemT));
    anItem -> key = (timeType) T + j;
    anItem -> id = curIndex;
    hInsert(h, anItem);
    printf("ins "); 
    fprintfll(stdout, "%1", (timeType) anItem->key);
    printf(" %d\n", curIndex);
    N--;
    checkIfDone;
  }

  nextItem = (itemT *) calloc(1, sizeof(itemT));

  while (1) { /* checkIfDone will exit */

    /* insert */
    j = Random(0,E);
    curIndex++;
    anItem = nextItem;
    anItem -> key = (timeType) T + j;
    anItem -> id = curIndex;
    hInsert(h, anItem);
    printf("ins "); 
    fprintfll(stdout, "%1", (timeType) anItem->key);
    printf(" %d\n", curIndex);
    N--;
    checkIfDone;

    /* extract min */
    printf("dmn\n");
    extractMin(h, nextItem);
    N--;
    checkIfDone;

    curT = nextItem->key;
    if (curT >= T)
      T += step;
  }

  exit (0);
}

