/* 
   Generator of monotone sequences of priority queue operations.
   Written by Andrew V. Goldberg, NEC Research Inst.,  avg@research.nj.nec.com.
   For format details, see the 5th DIMACS Implementation Challenge info.

*/
 
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <values.h>
#include <math.h>

#include "random.h"
#include "fprintfll.h"

#define RANGE 100
#define DASH '-'
#define OUT -1

/* types */
/*
  use long long int time to prevent overflow
  if long long not supported, use double
*/
typedef long long int timeType;

typedef struct itemSt
{
  long        heap_pos;  /* position in the heap */
  long        S_pos;     /* position in S */
  timeType    key;
  long        id;
} itemT;

typedef struct heapSt
{
  long    size;
  itemT **item;
} heap;

/* macros */
#define checkIfDone {if (N <= 0) exit(0);}

#include "heap.h"

void error (int errorNo)

{
  switch ( errorNo ) {

  case 1: {
    fprintf ( stderr, "\nUsage: mongen C K D N [-sS]\n");
    fprintf ( stderr, "      C is an upper bound on event duration \n");
    fprintf ( stderr, "      K is an approximate queue size\n");
    fprintf ( stderr, "      D is an approximate number of decrease keys per element\n");
    fprintf ( stderr, "      N is the total number of operations\n");
    fprintf ( stderr, "      S is a seed\n");
    break;
  }
  case 2: {
    fprintf ( stderr, "max duration C must be 2 or greater\n");
    break;
  }
  case 3: {
    fprintf ( stderr, "approx size K must be positive\n");
    break;
  }
  case 4: {
    fprintf ( stderr, "number of decrease keys D must be nonnegative\n");
    break;
  }
  case 5: {
    fprintf ( stderr, "number of operations N must be nonnegative\n");
    break;
  }
  }

  exit(errorNo);
}

void main (int argc, char* argv[])
{

  char args[30];
  long C, C2, D, K, N;
  long maxIndex;
  long i, j;
  long inc;
  long seed=31415;
  long curIndex=0;
  timeType T1, T2, curT;
  itemT **S;       /* array containing a set of items that can be decreased */
  long lastInS= -1;
  itemT *anItem, *nextItem;
  heap h;

  if ((argc < 5) || (argc > 6)) error(1);

  /* first parameter -- C */
  strcpy (args, argv[1]);
  if (( C = atoi(argv[1])) < 2) error (2);
 
  /* second parameter -- K */
  strcpy (args, argv[2]);
  if (( K = atoi(argv[2])) < 1) error (3);

  /* third parameter -- D */
  strcpy (args, argv[3]);
  if (( D = atoi(argv[3])) < 0) error (4);
 
  /* fourth parameter -- N */
  strcpy (args, argv[4]);
  if (( N = atoi(argv[4])) < 0) error (5);

  if ( argc == 6 ) {
    strcpy ( args, argv[5]);
    if (( args[0] != DASH ) || ( args[1] != 's')) error (1);
    seed  =  atoi ( &args[2] );
  }

  /* initialization */
  C2 = C/2;
  SetRandom(seed);
  S = (itemT **) calloc(K+1, sizeof(itemT *));
  makeHeap(h, K+1);
  curT = 0;
  T1 = C2;
  T2 = T1 + C2;

  /* figure out the number of insert operations, upper bound */
  maxIndex = 2 + K + (N / (D+2));

  printf("pqh %d -M%d -c%d -C%d\n", maxIndex, K+1, 0, C);
  printf("com mongen parameters C=%d K=%d D=%d N=%d\n",
	 C, K, D, N);

  checkIfDone;

  /* insert K elements */
  for (i = 0; i < K; i++) {
    
    curIndex++;
    anItem = (itemT *) calloc(1, sizeof(itemT));
    anItem -> key = T1;
    anItem -> id = curIndex;
    S[lastInS] = anItem;
    anItem->S_pos = lastInS;

    printf("ins "); 
    fprintfll(stdout, "%1", anItem->key);
    printf(" %d\n", curIndex);

    hInsert(h, anItem);
    lastInS++;
    N--;
    checkIfDone;
  }

  nextItem = (itemT *) calloc(1, sizeof(itemT));

  /* do insert--decrease-key--delete cycle */
  while (1) { /* checkIfDone will exit */
    for (i = 0; i < D; i++) {

      /* insert */
      curIndex++;
      anItem = nextItem;
      anItem -> key = T2;
      anItem -> id = curIndex;
      S[lastInS] = anItem;
      anItem->S_pos = lastInS;
      printf("ins "); 
      fprintfll(stdout, "%1", anItem->key);
      printf(" %d\n", curIndex);
      hInsert(h, anItem);
      lastInS++;
      N--;
      checkIfDone;

      /* decrease-key */
      /* chose random element in S */
      for (i = 1; i <= D; i++) {
	while (lastInS >= 0) {
	  j = Random(0,lastInS-1);
	  if (S[j]->key > curT) break;
	  else {
	    /* delete j */
	    lastInS--;
	    S[j]->S_pos = OUT;
	    if (j != lastInS) {
	      S[j] = S[lastInS];
	      S[j]->S_pos = j;
	    }
	  }
	}
	if (lastInS <= 0) {
	  break;
	}
	/* select decrease value */
	inc = Random(0, (long) sqrt((double) (S[j]->key - curT)));
	if (inc == 0) inc = 1;
	S[j]->key = S[j]->key - inc;
	printf("dcr "); 
	fprintfll(stdout, "%1", S[j]->key);
	printf(" %d\n", S[j]->id);
	decreaseKey(h, S[j], S[j]->key);
	N--;
	checkIfDone;
      }
      
      /* extract-min */
      printf("dmn\n");
      extractMin(h, nextItem);
      if (nextItem->S_pos != OUT) {
	lastInS--;
	S[lastInS]->S_pos = nextItem->S_pos;
	S[nextItem->S_pos] = S[lastInS];
	nextItem->S_pos = OUT;
      }
      curT = nextItem->key;
      if (curT > T1) {
	T1 = T2;
	T2 += C2;
      }
      N--;
      checkIfDone;
    }
  }

  exit (0);
}

